package org.netkernel.mod.architecture.bits;

import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.netkernel.ext.system.representation.*;
import org.netkernel.ext.system.representation.impl.ProtoMap;
import org.netkernel.ext.system.representation.impl.RepresentationReferences;
import org.netkernel.ext.system.representation.impl.SpaceGraph;
import org.netkernel.ext.system.representation.impl.SpaceGraph.Edge;
import org.netkernel.ext.system.transrepresentation.MetaToXML;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.boot.ModuleManager;
import org.netkernel.layer0.boot.ModuleSourceWithLevel;
import org.netkernel.layer0.meta.*;
import org.netkernel.mod.architecture.ISERep;
import org.netkernel.mod.architecture.Utils;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.util.*;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.*;
import org.netkernel.xml.xda.DOMXDA;
import org.netkernel.xml.xda.IXDAIterator;
import org.netkernel.xml.xda.XPathLocationException;
import org.w3c.dom.*;


public class EntityDetailsAccessor extends StandardAccessorImpl
{
	public void onSource(INKFRequestContext context) throws Exception
	{
		
		INKFRequestReadOnly reqIn=context.getThisRequest();
		String hash=reqIn.getArgumentValue("entityId");
		boolean references=reqIn.argumentExists("ref");

		String icon=null;
		String name=null;
		String desc="";
		Object rep=null;

		if (hash!=null)
		{	//generate detail panel
			String moduleHash=null, spaceHash=null, elementHash=null;
			int i1=hash.indexOf('/');
			if (i1>=0)
			{	moduleHash=hash.substring(0,i1);
				int i2=hash.indexOf('/',i1+1);
				if (i2>0)
				{	spaceHash=hash.substring(i1+1,i2);
					elementHash=hash.substring(i2+1);
				}
				else
				{	spaceHash=hash.substring(i1+1);
				}
			}
			else
			{	moduleHash=hash;
			}
			
			IRepDeployedModules deployedModules=(IRepDeployedModules)context.source("active:moduleList",IRepDeployedModules.class);
			String resourceSpaceHash=getResourceSpaceHash(moduleHash, spaceHash,deployedModules);
			IMetaRepresentation meta = deployedModules.metadataForHash(hash);
			if (meta==null)
			{	throw new NKFException("Explorer entity not found",hash);
			}
			
			int k=hash.lastIndexOf('/');
			String parentHash=(k>=0)?hash.substring(0,k):"";
			Document metaDoc=(Document)context.transrept(meta, Document.class);  
			
			String pageId="ed"+org.netkernel.mod.architecture.Utils.getUniqueId();

			context.setCWU(context.getRuntimeSourceIdentifier());
			INKFRequest req = context.createRequest("active:xslt");
			req.addArgument("operator","res:/org/netkernel/mod/architecture/bits/styleEntityDetails.xsl");
			req.addArgumentByValue("resourceSpaceHash", resourceSpaceHash);
			req.addArgumentByValue("parentHash", parentHash);
			req.addArgumentByValue("hash", hash);
			req.addArgumentByValue("pageId", pageId);
			req.addArgumentByValue("operand",metaDoc);
			req.addArgument("deployed","active:moduleListDoc");
			
			
			//search for any documentation
			if (elementHash!=null && !references)
			{
				INKFRequest req3=context.createRequest("active:nkseSearch");
				req3.addArgumentByValue("operand","eid:\""+elementHash+"\"");
				req3.addArgumentByValue("analyzer","org.apache.lucene.analysis.WhitespaceAnalyzer");
				req3.setRepresentationClass(IHDSNode.class);
				IHDSNode searchResult=(IHDSNode)context.issueRequest(req3);
				if (searchResult!=null)
				{	
					IHDSNodeList fields=searchResult.getNodes("/hits/hit/field");
					String documentID=(String)fields.filter(HDSPredicateFactory.namedChildStringEquals("name", "documentID")).getFirstValue("value");
					if (documentID!=null)
					{	//use existing documentation
						req3=context.createRequest("active:documentationEntry");
						req3.addArgument("id",documentID);
						Object docentry=context.issueRequest(req3);
						req3=context.createRequest("active:docRendererInner");
						req3.addArgumentByValue("docentry",docentry);
						req3.setRepresentationClass(String.class);
						String doc=(String)context.issueRequest(req3);
						req.addArgumentByValue("hrd",doc);
					}
					else
					{	//generate a dynamic reference doc
						String docsrc="{endpoint}"+hash+"{/endpoint}";
						req3=context.createRequest("active:rawDocRenderer");
						req3.addArgumentByValue("docsrc",docsrc);
						req3.setRepresentationClass(String.class);
						String doc=(String)context.issueRequest(req3);
						req.addArgumentByValue("hrd",doc);
					}
				}
			}
			
			//prototype stuff
			String protoHash=null;
			IPrototypeMeta protoMeta=null;
			if (meta instanceof IPhysicalEndpointMeta)
			{	Class<IEndpoint> impl=((IPhysicalEndpointMeta)meta).getImplementation();
				ProtoMap protoMap=context.source("active:moduleListProtoMap",ProtoMap.class);
				protoHash=protoMap.getPrototypeHashFor(impl);
				if (protoHash!=null)
				{	protoMeta=(IPrototypeMeta)deployedModules.metadataForHash(protoHash);
					HDSBuilder b = new HDSBuilder();
					b.pushNode("proto");
					b.addNode("name",protoMeta.getName());
					b.addNode("hash",protoHash);
					req.addArgumentByValue("proto", b.getRoot());
					
					//see if we have any customised state viewer this this prototype instance
					ISERep config = context.source("active:seConfig",ISERep.class);
					String viewId=null;
					String implId=protoMeta.getIdentifier().toString();
					IHDSNodeList views=config.getViewsFor("prototypeInstanceState");
					for (IHDSNode view: views)
					{	String prototype=(String)view.getFirstValue("prototype");
						if (prototype.equals(implId))
						{	viewId=(String)view.getFirstValue("id");
							break;
						}
					}
					if (viewId!=null)
					{	INKFResponseReadOnly viewResponse=config.generateView(viewId, hash, context);
						req.addArgumentFromResponse("statePlugin", viewResponse);
					}
					
					if (IStandardOverlay.class.isAssignableFrom(impl))
					{	ISpace delegateSpace=(ISpace)meta.getAdditionalFields().getValue(IStandardOverlay.STATE_DELEGATE_SPACE);
						String delegateSpaceHash=deployedModules.hashForComponent(delegateSpace);
						HDSBuilder b2=new HDSBuilder();
						b2.pushNode("delegateSpace");
						b2.addNode("name", delegateSpace.toString());
						b2.addNode("hash",delegateSpaceHash);
						req.addArgumentByValue("delegateSpace", b2.getRoot());
					}
					
				}
			}
			else if (references && meta instanceof IPrototypeMeta)
			{
				ProtoMap protoMap=context.source("active:moduleListProtoMap",ProtoMap.class);
				Class<IEndpoint> implClass=((IPrototypeMeta)meta).getImplementation();
				String locatedHash=protoMap.getPrototypeHashFor(implClass);
				String[] epInstanceHashes = protoMap.getEndpointInstancesOf(locatedHash);
				HDSBuilder b = new HDSBuilder();
				b.pushNode("refs");
				String lastModHash="";
				for (int i=0; i<epInstanceHashes.length; i++)
				{
					String instHash=epInstanceHashes[i];
					int j=instHash.indexOf('/');
					String modHash=instHash.substring(0,j);
					int j2=instHash.lastIndexOf('/');
					String sHash=instHash.substring(0,j2);

					b.pushNode("ref");
					b.addNode("modHash",modHash);
					IMetaRepresentation modMeta=deployedModules.metadataForHash(modHash);
					b.addNode("mod", modMeta.getName());
					b.addNode("spHash",sHash);
					IMetaRepresentation spaceMeta=deployedModules.metadataForHash(sHash);
					b.addNode("sp", spaceMeta.getName());
					b.addNode("epHash",instHash);
					Object instance=deployedModules.componentForHash(instHash);
					b.addNode("ep", instance.toString());
					b.popNode();
					
					b.pushNode("inst");
					b.addNode("hash", instHash);
					Object inst=deployedModules.componentForHash(instHash);
					b.addNode("toString", inst.toString());
					b.popNode();
				}
				req.addArgumentByValue("refs", b.getRoot());
			}
			
			if (references && meta instanceof IRepresentationMeta)
			{
				String originPhysicalHash=Utils.findOriginatingPhysicalEndpoint(hash, deployedModules);
				int j=originPhysicalHash.lastIndexOf('/');
				String originSpaceHash=originPhysicalHash.substring(0,j+1);
				j=hash.lastIndexOf('/');
				String originRepHash=originSpaceHash+hash.substring(j+1);
				
				RepresentationReferences repMap=context.source("active:moduleListRepMap",RepresentationReferences.class);
				String[] repInstanceHashes = repMap.getReferencesTo(originRepHash);
				HDSBuilder b = new HDSBuilder();
				b.pushNode("refs");
				String lastModHash="";
				if (repInstanceHashes!=null)
				{
					for (int i=0; i<repInstanceHashes.length; i++)
					{
						String instHash=repInstanceHashes[i];
						j=instHash.indexOf('/');
						String modHash=instHash.substring(0,j);
						j=instHash.lastIndexOf('/');
						String sHash=instHash.substring(0,j);
	
						b.pushNode("ref");
						b.addNode("modHash",modHash);
						IMetaRepresentation modMeta=deployedModules.metadataForHash(modHash);
						b.addNode("mod", modMeta.getName());
						b.addNode("spHash",sHash);
						IMetaRepresentation spaceMeta=deployedModules.metadataForHash(sHash);
						b.addNode("sp", spaceMeta.getName());
						b.addNode("lepHash",instHash);
						IMetaRepresentation epMeta=deployedModules.metadataForHash(instHash);
						b.addNode("lep", epMeta.getName());
						b.popNode();
					}
				}
				req.addArgumentByValue("refs", b.getRoot());
			}
			
			
			
			//space references
			if (references && meta instanceof ISpaceMeta)
			{
				HDSBuilder b = new HDSBuilder();
				b.pushNode("refs");
				SpaceGraph sg=context.source("active:moduleListSpaceGraph",SpaceGraph.class);
				Edge[] edges=sg.getEdgesIn(hash);
				if (edges!=null)
				{	writeSpaceReferences(edges,b,deployedModules);
				}
				req.addArgumentByValue("refs", b.getRoot());
			}
			//module references
			if (meta instanceof IModuleMeta)
			{
				IModule module=(IModule)deployedModules.componentForHash(hash);
				HDSBuilder b = new HDSBuilder();
				b.pushNode("refs");
				
				ModuleSourceWithLevel mswl = ModuleManager.getSingleton().getRegisteredModules().get(module.getSource());
				if (mswl!=null)
				{	b.addNode("runlevel", mswl.getLevel());
					b.addNode("declarationtype", ModuleManager.getSingleton().getDeclarationTypeFor(mswl.getDeclaration()));
					b.addNode("declaration", mswl.getDeclaration());
				}
				
				
				if (references)
				{	SpaceGraph sg=context.source("active:moduleListSpaceGraph",SpaceGraph.class);
					if (module!=null)
					{	for (ISpace space: module.getPublicSpaces())
						{	String sh=deployedModules.hashForComponent(space);
							Edge[] edges=sg.getEdgesIn(sh);
							if (edges!=null)
							{	writeSpaceReferences(edges,b,deployedModules);
							}
						}
					}
				}
				req.addArgumentByValue("refs", b.getRoot());
			}
			
			rep=context.issueRequest(req);
			
			req=context.createRequest("active:xrl2");
			req.addArgumentByValue("template",rep);
			rep=context.issueRequest(req);
			
			DOMXDA metaXDA = new DOMXDA(metaDoc,false);
			try
			{	String iconPath=metaXDA.getText("/meta/icon",true);
				if (iconPath.length()>0)
				{	icon="/tools/cr/"+resourceSpaceHash+"/"+iconPath;
				}
			}
			catch (XPathLocationException e) {;}
			if (icon==null)
			{	if (meta instanceof IRepresentationMeta)
				{	icon="/tools/ae/pub/images/48/Representation48.png";
				}
				else if (meta instanceof IPrototypeMeta)
				{	icon="/tools/ae/pub/images/48/Prototype48.png";
				}
				else if (meta instanceof IEndpointMeta)
				{	icon="/tools/ae/pub/images/48/Endpoint48.png";
				}
				else if (meta instanceof ISpaceMeta)
				{	icon="/tools/ae/pub/images/48/Space48.png";
				}
				else if (meta instanceof IModuleMeta)
				{	icon="/tools/ae/pub/images/48/Module48.png";
				}
				else if (meta instanceof IPhysicalEndpointMeta)
				{	
					if (protoMeta!=null)
					{	IIdentifier protoIcon=protoMeta.getIcon();
						if (protoIcon!=null)
						{	int i=protoHash.lastIndexOf('/');
							resourceSpaceHash=protoHash.substring(0,i);
							icon="/tools/cr/"+resourceSpaceHash+"/"+protoIcon.toString();
						}
					}
					if (icon==null)
					{	icon="/tools/ae/pub/images/48/Physical48.png";
					}					
				}
			}
			name=metaXDA.getText("/meta/name", true);
			try
			{	desc=metaXDA.getText("/meta/desc", true);
			}
			catch (XPathLocationException e) {;}
		}

		INKFResponse resp=context.createResponseFrom(rep);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
		resp.setHeader("WrappedControlPanel", name);
		resp.setHeader("WrappedControlPanelSubtitle", desc);
		if (icon!=null)
		{	resp.setHeader("WrappedControlPanelIcon", icon);
		}
		
	}
	
	private void writeSpaceReferences(Edge[] edges, HDSBuilder b, IRepDeployedModules deployedModules)
	{
		for (Edge edge: edges)
		{	String epHash=edge.getPhysicalEndpointHash();
			String spHash=edge.getSpaceHash();
			String modHash=spHash.substring(0,spHash.indexOf('/'));
			Object ep=deployedModules.componentForHash(epHash);
			IMetaRepresentation sp=deployedModules.metadataForHash(spHash);
			IMetaRepresentation mod=deployedModules.metadataForHash(modHash);
			b.pushNode("ref");
			b.addNode("epHash", epHash);
			b.addNode("ep", ep.toString());
			b.addNode("spHash", spHash);
			b.addNode("sp", sp.getName());
			b.addNode("modHash", modHash);
			b.addNode("mod", mod.getName());
			b.popNode();
		}
	}
	
	private String getResourceSpaceHash(String aModule, String aSpace, IRepDeployedModules aDeployedModules)
	{
		String result;
		if (aSpace!=null)
		{	result=aModule+"/"+aSpace;
		}
		else
		{	// find first public space in module
			result="none";
			IModule module=(IModule)aDeployedModules.componentForHash(aModule);
			if (module!=null)
			{	ISpace[] spaces = module.getPublicSpaces();
				if (spaces.length>0)
				{	result = aDeployedModules.hashForComponent(spaces[0]);
				}
			}
		}
		return result;
	}
	
	/*
	private INKFResponseReadOnly<ExplorerPluginsRepresentation> mPluginsResponse;
	
	private ExplorerPluginsRepresentation getPlugins(INKFRequestContext aContext) throws Exception
	{	INKFResponseReadOnly<ExplorerPluginsRepresentation> lastPluginsResponse=mPluginsResponse;
		if (lastPluginsResponse==null || lastPluginsResponse.isExpired())
		{	lastPluginsResponse=aContext.sourceForResponse("active:explorerPlugins",ExplorerPluginsRepresentation.class);
		}
		return lastPluginsResponse.getRepresentation();
	}
	*/
	
	/*
			// append any documentation
			IIdentifier id = meta.getHumanReadableDocumentation();
			if (id!=null && resourceSpaceHash!=null)
			{	
				try
				{
					INKFRequest req2 = context.createRequest(id.toString());
					ISpace resourceSpace = (ISpace)deployedModules.componentForHash(resourceSpaceHash);
					
					IRequestScopeLevel hrdRctx = RequestScopeLevelImpl.createNonDurableScopeLevel(resourceSpace,context.getKernelContext().getThisKernelRequest().getRequestScope());
					req2.setRequestScope(hrdRctx);
					req2.setRepresentationClass(String.class);
					INKFResponseReadOnly hrdResp=context.issueRequestForResponse(req2);
					String hrdString;
					
					INKFRequest docReq=context.createRequest("active:wikiParser/XHTML");
					docReq.addArgumentByValue("operand",hrdResp.getRepresentation());
					docReq.setRepresentationClass(String.class);
					docReq.addArgument("base","/");
					hrdString=(String)context.issueRequest(docReq);
					
					//Do the book cross linking (sync)
					Map bookmap=(Map)context.source("active:documentBookMap");
					Pattern mPattern=Pattern.compile("./wiki/(.*?)\\\"");
					Matcher m=mPattern.matcher(hrdString);
					int i=1;
					int start=0;
					while(m.find(start))
					{	String doc2=m.group(1);
						String sub=null;
						hash=deployedModules.hashForSpaceElement(doc2);
						if(hash!=null)
						{	sub="/tools/explore2/"+hash+"\"";
						}
						else
						{	String lbookid=(String)bookmap.get(doc2);
							if(lbookid==null)
							{	sub="./"+doc2+"\"";
							}
							else
							{	sub="../"+lbookid+"/"+doc2+"\"";;		
							}
						}
						hrdString=m.replaceFirst(sub);
						start=m.start();
						m.reset(hrdString);
					}
					req.addArgumentByValue("hrd",hrdString);
				}
				catch (Exception e)
				{	// failed to find doc
					System.out.println(org.netkernel.layer0.util.Utils.throwableToString(e));
				}
			}
	*/
	
	
}
