package org.netkernel.mod.architecture.bits.channel;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.ext.system.representation.impl.SpaceGraph;
import org.netkernel.ext.system.representation.impl.SpaceGraph.Edge;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.builtin.branchmerge.BranchMergeSpace;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

import java.awt.Color;
import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.util.*;
import java.util.Map.Entry;

public class ChannelViewAccessor extends StandardAccessorImpl
{
	public ChannelViewAccessor()
	{	this.declareThreadSafe();
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		String entityHash=aContext.getThisRequest().getArgumentValue("entityId");
		IRepDeployedModules modules=aContext.source("active:moduleList",IRepDeployedModules.class);
		
		IHDSNodeList allModules=modules.getHierarchy().getNodes("/modules/module");
		IHDSNode ourModule=allModules.filter(HDSPredicateFactory.namedChildStringEquals("hash", entityHash)).getFirstNode();
		//System.out.println(ourModule);
		
		
		IHDSNodeList spaceNodes=ourModule.getNodes("//spaces/space");
		System.out.println("spaces="+spaceNodes.size());
		
		//find all logical endpoints and their origin space
		Map<String,LogicalEndpointDefinition> logicalEndpoints=new HashMap<String, LogicalEndpointDefinition>();
		for (IHDSNode spaceNode : spaceNodes)
		{
			String spaceHash=(String)spaceNode.getFirstValue("hash");
			ISpaceMeta spaceMeta=(ISpaceMeta)modules.metadataForHash(spaceHash);
			ISpaceElements spaceElements=spaceMeta.getElements();
			for (int i=0; i<spaceElements.size(); i++)
			{
				IIdentifier id=spaceElements.getIdentifier(i);
				if (id!=null)
				{
					String ids=id.toString();
					LogicalEndpointDefinition led=logicalEndpoints.get(ids);
					if (led==null)
					{	led=new LogicalEndpointDefinition(ids);
						logicalEndpoints.put(ids, led);
					}
					ISpace delegatedSpace=spaceElements.getDelegatedSpace(i);
					String delegatedHash=null;
					if (delegatedSpace!=null)
					{	delegatedHash=modules.hashForComponent(delegatedSpace);
					}
					IEndpoint physicalEp=spaceElements.getPhysicalEndpoint(i);
					String physicalHash=modules.hashForComponent(physicalEp);
					
					led.declareSpace(spaceHash,physicalHash,delegatedHash,modules);
				}
			}
			
		}
		System.out.println("leds="+logicalEndpoints.size());
		
		//remove logical endpoints which originate outside our target module(s)
		String ehHead=entityHash+"/";
		for (Iterator<Entry<String,LogicalEndpointDefinition>> i=logicalEndpoints.entrySet().iterator(); i.hasNext(); )
		{	Entry<String,LogicalEndpointDefinition> e=i.next();
			LogicalEndpointDefinition led=e.getValue();
			
			String originHash=led.getOrigin();
			if (!originHash.startsWith(ehHead))
			{	i.remove();
			}
		}
		
		System.out.println("local leds="+logicalEndpoints.size());
		
		//merge logical endpoints into logical channels (follow the same routing)
		Map<String,ChannelDefinition> channels = new LinkedHashMap<String, ChannelDefinition>();
		for (IHDSNode spaceNode : spaceNodes)
		{
			String spaceHash=(String)spaceNode.getFirstValue("hash");
			ISpace space = (ISpace)modules.componentForHash(spaceHash);
			if (!(space instanceof BranchMergeSpace))
			{
				for (Iterator<Entry<String,LogicalEndpointDefinition>> i=logicalEndpoints.entrySet().iterator(); i.hasNext(); )
				{	Entry<String,LogicalEndpointDefinition> e=i.next();
					LogicalEndpointDefinition led=e.getValue();
					if (led.startsIn(spaceHash))
					{
						String pathString=led.getPathStringFrom(spaceHash);
						ChannelDefinition cd=channels.get(pathString);
						if (cd==null)
						{	cd=new ChannelDefinition(pathString, led.getPathFrom(spaceHash));
							channels.put(pathString, cd);
						}
						cd.addLogicalEndpoint(e.getKey());
					}
				}
			}
		}
		
		//debug
		System.out.println("channels="+channels.size());
		/*
		for (ChannelDefinition cd : channels.values())
		{	if (cd.getPathList().size()>1)
			{	System.out.println("channel: "+cd.getEndpoints().size()+" "+cd.getPath());
			}
		}
		*/
		
		SpaceGraph sg = aContext.source("active:moduleListSpaceGraph", SpaceGraph.class);

		//remove channels that don't go anywhere
		for (Iterator<Entry<String,ChannelDefinition>> i=channels.entrySet().iterator(); i.hasNext(); )
		{	Entry<String,ChannelDefinition> e=i.next();
			if (e.getValue().getPathList().size()<2)
			{	//i.remove();
				String orphanedSpaceHash=e.getValue().getPathList().get(0);
				
				Edge[] edges=sg.getEdgesIn(orphanedSpaceHash);
				if (edges!=null && edges.length==1)
				{
					//System.out.println("orphaned "+orphanedSpaceHash+" from "+edges.length);
					String surrogateParentSpaceHash=edges[0].getSpaceHash();
					e.getValue().addOrphanParent(surrogateParentSpaceHash);
				}
				else
				{	//must be injected space
					i.remove();
				}
			}
		}
		
		//build depth tree for spaces involved in diagram and put channels into spaces
		Map<String,SpaceDefinition> spaceDefinitions=new HashMap<String, SpaceDefinition>();
		for (ChannelDefinition cd : channels.values())
		{	
			int depth=0;
			SpaceDefinition parent=null;
			for (int i=cd.getPathList().size()-1; i>=0; i--)
			{	String spaceHash=cd.getPathList().get(i);
				SpaceDefinition sd=spaceDefinitions.get(spaceHash);
				
				if (sd==null)
				{	sd=new SpaceDefinition(spaceHash, depth);
					spaceDefinitions.put(spaceHash, sd);
				}
				else
				{	
					if (sd.getDepth()<=depth)
					{	depth=sd.getDepth();
					}
					else
					{	//shuffle down this and all dependencies
						sd.pushDepth(depth);
					}
				}
				if (parent!=null)
				{	parent.addDependency(sd);
				}
				sd.addChannel(cd);
				depth--;
				parent=sd;
			}
			
		}
		
		//calculate min depth to offset back to zero
		int minDepth=0;
		for (SpaceDefinition sd: spaceDefinitions.values())
		{	int depth=sd.getDepth();
			if (depth<minDepth) minDepth=depth;
		}
		
		//offset depths
		for (SpaceDefinition sd: spaceDefinitions.values())
		{	sd.offsetDepth(-minDepth);
			//System.out.println(sd.getDepth()+" "+sd.getChannels().size()+" "+modules.componentForHash(sd.getSpaceHash()));
		}
		
		int maxDepth=-minDepth;
		
		//look from permutations of channels by shuffling
		int[] p=new int[channels.size()];
		for (int i=0; i<p.length; i++)
		{	p[i]=i;
		}
		Shuffle initialShuffle=new Shuffle(p);
		List<ChannelDefinition> channelList=new ArrayList<ChannelDefinition>(channels.values());
		List<Shuffle> results=new ArrayList<Shuffle>();
		recurseShuffle(channelList,spaceDefinitions,initialShuffle,maxDepth,0,results);

		//choose best shuffle
		Shuffle best=initialShuffle;
		for (Shuffle shuffle: results)
		{	if (shuffle.getBadness()<best.getBadness())
			{	best=shuffle;
			}
		}
		System.out.println(best.toString()+" (best)");
		List<ChannelDefinition> channelList2=new ArrayList<ChannelDefinition>(channelList.size());
		for (int i=0; i<channelList.size(); i++)
		{	channelList2.add(channelList.get(best.get(i)));
		}
		
		//generate columns for choosen shuffle
		List<Column> columns=getColumnsForShuffle(best, channelList, spaceDefinitions, maxDepth);
		
		
		//create grid of channels and spaces
		
		IHDSNode diagram=generateDiagram(spaceDefinitions,channelList2,columns,modules,spaceNodes,aContext);
		INKFResponse resp=aContext.createResponseFrom(diagram);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
		resp.setMimeType("image/svg+xml");
	}
	
	private void recurseShuffle(List<ChannelDefinition> aChannels, Map<String,SpaceDefinition> aSpaceDefinitions, Shuffle aShuffle, int aMaxDepth, int aRecurse, List<Shuffle> aResults)
	{
		List<Column> columns=getColumnsForShuffle(aShuffle, aChannels, aSpaceDefinitions, aMaxDepth);
		
		float totalBadness=0;
		Column baddest1=null;
		Column baddest2=null;
		float baddestValue=0.0f;
		float f=1.0f;
		for (Column col : columns)
		{	float badness=f*col.getBadness();
			totalBadness+=badness;
			f=f*0.9f;
			if (badness>baddestValue)
			{	baddestValue=badness;
				baddest1=col;
				baddest2=null;
			}
			else
			{	if (badness==baddestValue)
				{	//System.out.println("here");
					baddest2=col;
				}
			}
		}
		
		aShuffle.setBadness(totalBadness);
		//System.out.println(aShuffle.toString());
		aResults.add(aShuffle);
		
		
		if (aRecurse<6 && totalBadness>0.0f)
		{
			for (Shuffle shuffle : baddest1.getShuffles(aShuffle))
			{	recurseShuffle(aChannels, aSpaceDefinitions, shuffle, aMaxDepth, aRecurse+1,aResults);
			}
			if (baddest2!=null)
			{
				for (Shuffle shuffle : baddest2.getShuffles(aShuffle))
				{	recurseShuffle(aChannels, aSpaceDefinitions, shuffle, aMaxDepth, aRecurse+1,aResults);
				}
			}
		}
	}
	
	private List<Column> getColumnsForShuffle(Shuffle aShuffle, List<ChannelDefinition> aChannels, Map<String,SpaceDefinition> aSpaceDefinitions, int aMaxDepth)
	{	List<Column> columns=new ArrayList<Column>();
		for (int depth=0; depth<=aMaxDepth; depth++)
		{	SpaceDefinition lastSpace=null;
			SpaceGroup lastGroup=null;
			Map<String,SpaceGroup> groups=new HashMap<String,SpaceGroup>();
			List<SpaceGroup> groupList=new ArrayList<SpaceGroup>();
			int c=0;
			for (int j=0; j<aChannels.size(); j++)
			{	ChannelDefinition cd=aChannels.get(aShuffle.get(j));
				//find space in this location
				SpaceDefinition sd=null;
				boolean hasChannelThrough=false;
				for (int i=cd.getPathList().size()-1; i>=0; i--)
				{	String spaceHash=cd.getPathList().get(i);
					SpaceDefinition sd2=aSpaceDefinitions.get(spaceHash);
					if (sd2.getDepth()==depth)
					{	sd=sd2;
						//break;
					}
					else if (sd2.getDepth()>depth)
					{	hasChannelThrough=true;
						//break;
					}
				}
				
				if (lastSpace!=sd)
				{	if (lastGroup!=null)
					{	lastGroup.setEnd(c);
					}
					SpaceGroup newGroup=new SpaceGroup(sd, c);
					if (sd==null && hasChannelThrough) newGroup.setHasChannelThrough();
					groupList.add(newGroup);
					lastGroup=newGroup;
					lastSpace=sd;
				}
				c++;
			}
			if (lastGroup.getSpace()==null)
			{	groupList.remove(lastGroup);
			}
			else
			{	lastGroup.setEnd(aChannels.size());
			}
			
			Column col=new Column(depth,groupList);
			columns.add(col);
		}
		return columns;
	}
	
	static Shuffle shuffleChannels(Shuffle aShuffle, int aSrcStart, int aSrcEnd, int aDest)
	{
		int[] result=new int[aShuffle.size()];
		int p=0;
		if (aDest<aSrcStart)
		{	for (int i=0; i<aDest; i++)
			{	result[p++]=aShuffle.get(i);
			}
			for (int i=aSrcStart; i<aSrcEnd; i++)
			{	result[p++]=aShuffle.get(i);
			}
			for (int i=aDest; i<aSrcStart; i++)
			{	result[p++]=aShuffle.get(i);
			}
			for (int i=aSrcEnd; i<aShuffle.size(); i++)
			{	result[p++]=aShuffle.get(i);
			}
		}
		else
		{	for (int i=0; i<aSrcStart; i++)
			{	result[p++]=aShuffle.get(i);
			}
			for (int i=aSrcEnd; i<aDest; i++)
			{	result[p++]=aShuffle.get(i);
			}
			for (int i=aSrcStart; i<aSrcEnd; i++)
			{	result[p++]=aShuffle.get(i);
			}
			for (int i=aDest; i<aShuffle.size(); i++)
			{	result[p++]=aShuffle.get(i);
			}
		}
		
		return new Shuffle(result);
	}
	
	
	
	private IHDSNode generateDiagram(Map<String,SpaceDefinition> aSpaces, List<ChannelDefinition> aChannels, List<Column> aColumns, IRepDeployedModules aModules, IHDSNodeList aSpaceNodes, INKFRequestContext aContext) throws Exception
	{
		float h=1600.0f;
		float w=1024.0f;
		
		HDSBuilder b=new HDSBuilder();
		b.pushNode("svg");
		b.addNode("@xmlns","http://www.w3.org/2000/svg");
		b.addNode("@version", "1.1");
		b.addNode("@width",(int)w);
		b.addNode("@height",(int)h);
		
		b.pushNode("style", aContext.source("res:/org/netkernel/mod/architecture/bits/channel/style.css",String.class));
		b.addNode("@type","text/css");
		b.popNode();

		b.pushNode("g");
		//b.addNode("@transform", "scale("+w+" "+h+")");
		
		b.pushNode("rect");
		b.addNode("@x","0");
		b.addNode("@y","0");
		b.addNode("@width",w);
		b.addNode("@height",h);
		b.addNode("@class","bg");
		b.popNode();
		
		int xs=40;
		int ys=10;
		int xo=350;
		int yo=ys/4;
		int minHeightForSpace=8;
		
		//generate heights for each channel
		for (ChannelDefinition cd : aChannels)
		{	String title=cd.getTitle(aModules);
			List<String> titleRows=Utils.splitIntoRows(title, 60, " /-");
			int height=3+titleRows.size();
			if (height<minHeightForSpace)
			{
				//check to see if channel contains any spaces used only by this channel
				boolean exclusive=false;
				for (String spaceHash: cd.getPathList())
				{	if (aSpaces.get(spaceHash).getChannels().size()==1)
					{	exclusive=true;
						break;
					}
				}
				if (exclusive)
				{	height=minHeightForSpace;
				}
			}
			cd.setHeight(height);
		}
		
		//find channels with branches
		
		
		String[] palette;
		//palette=new String[]{  "D9CEB2", "948C75", "D5DED9", "7A6A53", "99B2B7" }; //good friends
		palette=new String[]{ "0834C3", "B163A3", "009B77", "DD4124", "D65076", "EFC050", "5B5EA6", "98B4D4", "DFCFBE" }; 
		
		//draw channels
		{
			int y=0;
			for (ChannelDefinition cd : aChannels)
			{	
				List<String> pathList=cd.getPathList();
				String spaceHash1=aSpaces.get(pathList.get(0)).getSpaceHash();
				
				// only when looking for physical endpoints
				//int j=spaceHash1.lastIndexOf('/');
				//spaceHash1=spaceHash1.substring(0, j);
				
				IHDSNode spaceNode=aSpaceNodes.filter(HDSPredicateFactory.namedChildStringEquals("hash", spaceHash1)).getFirstNode();
				boolean isPublic=spaceNode!=null && spaceNode.getFirstValue("public").toString().equals("true");
				//System.out.println(isPublic);
				
				int x0=aSpaces.get(pathList.get(0)).getDepth()*xs;
				int x1=isPublic?2:x0; //aSpaces.get(pathList.get(0)).getDepth();
				int x2=xo+aSpaces.get(pathList.get(pathList.size()-1)).getDepth()*xs+xs/2;

				int chTop=getCummulativeHeightOfChannel(y-1,aChannels);
				int chBot=getCummulativeHeightOfChannel(y,aChannels);
				
				b.pushNode("rect");
				b.addNode("@x",x1);
				b.addNode("@y",yo+chTop*ys+ys);
				b.addNode("@width",(x2-x1));
				b.addNode("@height",(chBot-chTop)*ys-ys*2);
				b.addNode("@rc",5);
				b.addNode("@ry",5);
				b.addNode("@class","channel");
				String rgb=palette[y%palette.length];
				b.addNode("@style","fill:#"+rgb);			
				b.popNode();
				
				boolean borw=getTextColorFor(rgb);
				
				
				String title=cd.getTitle(aModules);
				List<String> titleRows=Utils.splitIntoRows(title, 60, " /-");
				for (int i=0; i<titleRows.size(); i++)
				{
					b.pushNode("text",titleRows.get(i));
					b.addNode("@x",x1+5);
					b.addNode("@y",yo+chTop*ys+25+i*10);
					b.addNode("@font-size",10);
					//b.addNode("@font-weight","bold");
					b.addNode("@text-anchor","left");
					b.addNode("@fill",borw?"#000":"#FFF");
					b.popNode();
				}
				
				y++;
			}
		}

		//draw space groups
		for (Column col: aColumns)
		{
			Column colMerged=col.getWithGapsMerged();
			for (SpaceGroup sg: colMerged.getSpaceGroups())
			{
				if (sg.getSpace()!=null)
				{
					int x=xo+col.getDepth()*xs;
					//int y=yo+sg.getStart()*ys+ys/8;
					int chTop=getCummulativeHeightOfChannel(sg.getStart()-1,aChannels);
					int chBot=getCummulativeHeightOfChannel(sg.getEnd()-1,aChannels);
					int y=yo+chTop*ys+ys/2;
					int height=(chBot-chTop)*ys-ys;
					//int height=(sg.getEnd()-sg.getStart())*ys-ys/4;
					b.pushNode("rect");
					b.addNode("@x",x);
					b.addNode("@y",y);
					b.addNode("@rc",5);
					b.addNode("@ry",5);
					b.addNode("@width",xs/2);
					b.addNode("@height",height);
					b.addNode("@class","space");

					//String rgb=colourFromData(sg.getSpace().getSpaceHash().getBytes());
					//b.addNode("@style","fill:#"+rgb);				
					b.popNode();
					
					b.pushNode("g");
					b.addNode("@transform", "translate("+(x+xs/3)+" "+(y+height/2)+") rotate(-90)");
					
					String spaceHash=sg.getSpace().getSpaceHash();
					ISpaceMeta spaceMeta=(ISpaceMeta)aModules.metadataForHash(spaceHash);
					ISpaceElements el=spaceMeta.getElements();
					//sg.getSpace().
					
					
					//Object endpoint=aModules.componentForHash(spaceHash);
					//System.out.println("spaceHash="+spaceHash);
					//String text=endpoint!=null?endpoint.toString():"??";
					b.pushNode("text",spaceMeta.getName());
					b.addNode("@x",0);
					b.addNode("@y",0);
					b.addNode("@class","space-text");
					//b.addNode("@font-size",10);
					//b.addNode("@stroke",10);
					//b.addNode("@font-weight","bold");
					//b.addNode("@text-anchor","middle");
					b.popNode();
					
					b.popNode();
					
					
				}
			}
		}
				
		
		
		
		
		return b.getRoot();
		
	}
	
	private boolean getTextColorFor(String aRGB)
	{
		byte[] rgb=Utils.fromHexString(aRGB);
		int r=rgb[0]; if (r<0) r+=256;
		int g=rgb[1]; if (g<0) g+=256;
		int b=rgb[2]; if (b<0) b+=256;		
		float brightness=(0.2126f*r) + (0.7152f*g) + (0.0722f*b);		
		return brightness>128.0f;
	}
	
	private int getCummulativeHeightOfChannel(int aChannelIndex, List<ChannelDefinition> aChannels)
	{	int h=0;
		for (int i=0; i<=aChannelIndex; i++)
		{	h+=aChannels.get(i).getHeight();
		}
		return h;
	}
	
	
	private String colourFromData(byte[] aBytes) throws Exception
	{
		MessageDigest md=MessageDigest.getInstance("MD5");
		md.update(aBytes);
		byte[] digest=md.digest();
		
		String r=byteToHex(digest[0]);
		String g=byteToHex(digest[1]);
		String b=byteToHex(digest[2]);

		return r+g+b;
	}
	
	private String byteToHex(byte aByte)
	{
		int i=aByte;
		if (i<0) i+=256;
		i=128+i/2;
		String r=Integer.toHexString(i);
		if (r.length()==1) r="0"+r;
		return r;
	}
	
	
}
