package org.netkernel.mod.architecture.bits.staticstructure;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.ext.system.representation.impl.SpaceGraph;
import org.netkernel.ext.system.representation.impl.SpaceGraph.Edge;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.urii.IEndpoint;

public class GraphNode
{
	public float mX;
	public float mY;
	public Map<String,Float> mYComponentsByParent = new HashMap<String,Float>();
	public int mDepth;
	public float mXV;
	public float mYV;
	public List<GraphNode> mEdgesIn = new ArrayList<GraphNode>();
	public List<GraphNode> mEdgesOut = new ArrayList<GraphNode>();
	public String mHash;
	public boolean mExternal;
	public List<String> mPhysicalEndpoints;
	public List<GraphEdge> mNEdgesOut = new ArrayList<GraphEdge>();
	public boolean mHackFlag;
	
	private float mYBounds;
	private float mXBounds;
	
	public GraphNode(String aHash, float aX, boolean aExternal, IRepDeployedModules aModules, SpaceGraph aGraph)
	{	mHash=aHash;
		mX=aX;
		mExternal=aExternal;
		
		if (!mExternal)
		{	ISpaceMeta spaceMeta=(ISpaceMeta)aModules.metadataForHash(aHash);
			ISpaceElements spaceElements=spaceMeta.getElements();
			List<String> eps=new ArrayList<String>();
			IEndpoint lastEp=null;
			for (int i=0; i<spaceElements.size(); i++)
			{	IEndpoint ep=spaceElements.getPhysicalEndpoint(i);
				if (ep!=lastEp)
				{	String epHash=aModules.hashForComponent(ep);
					eps.add(epHash);
					lastEp=ep;
				}
			}
			mPhysicalEndpoints=eps;
		}
		else
		{	mPhysicalEndpoints=Collections.EMPTY_LIST;
		}
		
	}
	
	public float getXBounds()
	{	return mXBounds;
	}
	public float getYBounds()
	{	return mYBounds>getHeight()?mYBounds:getHeight();
	}
	public void setBounds(float aX, float aY)
	{	mXBounds=aX;
		mYBounds=aY;
	}
	
	public void updateYPos(float aYPos, String aParentHash)
	{
		
		mYComponentsByParent.put(aParentHash,aYPos);
		float ypos=0;
		for (Float v: mYComponentsByParent.values())
		{	ypos+=v;
		}
		ypos/=mYComponentsByParent.size();
		mY=ypos;
	}
	
	public void addNodeIn(GraphNode aNode)
	{	mEdgesIn.add(aNode);
	}
	
	public void addNodeOut(GraphNode aNode)
	{	mEdgesOut.add(aNode);
	}
	
	public boolean equals(Object aOther)
	{	return aOther instanceof GraphNode && ((GraphNode)aOther).mHash==mHash;
	}
	
	public int hashCode()
	{	return mHash.hashCode();
	}
	
	public int indexOfEndpoint(String aHash)
	{	int result=-1;
		for (int i=0; i<mPhysicalEndpoints.size(); i++)
		{	if (mPhysicalEndpoints.get(i).equals(aHash))
			{	result=i;
				break;
			}
		}
		return result;
	}
	
	public float getHeight()
	{	float h=(mPhysicalEndpoints.size()+2)*StaticStructureDiagramAccessor.ENDPOINT_HEIGHT;
		if (h<40.0f) h=40.0f;
		return h;
	}
	
	public float getYPosOfEndpoint(int aEndpointIndex)
	{	float h=getHeight();
		float y= ((aEndpointIndex+2)*StaticStructureDiagramAccessor.ENDPOINT_HEIGHT) - h/2;
		return y;
	}

	public int getYSpan()
	{
		return recurseGetYSpan(this);
	}
	
	private static int recurseGetYSpan(GraphNode aNode)
	{
		int h=0;
		if (aNode.mEdgesIn.size()<2)
		{
			for (GraphNode n: aNode.mEdgesOut)
			{	h+=recurseGetYSpan(n);
			}
			if (h==0)
			{	h=1;
			}
		}
		//System.out.println("GetYSpan: "+aNode.mHash+" "+h);
		return h;
	}
	
	
}
