package tools.logging;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

public class LogViewAccessor extends StandardAccessorImpl
{
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		IHDSNode param = aContext.source("httpRequest:/params",IHDSNode.class);
		IHDSNode logs=aContext.source("netkernel:/logLevels",IHDSNode.class);
		
		String logId=null;
		int records=20;
		int offset=0;
		if (param.getChildren().length==0)
		{	//initialise default values
			for (IHDSNode log: logs.getNodes("/logs/log"))
			{	if (log.getFirstValue("file")!=null)
				{	logId=log.getFirstValue("id").toString();
					break;
				}
			}
		}
		else
		{	try
			{	logId=param.getFirstValue("log").toString();
				records=Integer.parseInt(param.getFirstValue("records").toString());
				offset=Integer.parseInt(param.getFirstValue("offset").toString());
			}
			catch (Exception e) {;}
			
			if (param.getNodes("prev.x").size()>0)
			{	offset+=records;
			}
			else if (param.getNodes("next.x").size()>0)
			{	offset-=records;
			}
			else if (param.getNodes("end.x").size()>0)
			{	offset=0;
			}
		}
		
		HDSBuilder b = new HDSBuilder();
		b.addNode("log", logId);
		b.addNode("records", records);
		b.addNode("offset", offset);
		param=b.getRoot();
		
		INKFResponseReadOnly logRecords=null;
		if (logId!=null)
		{	INKFRequest req=aContext.createRequestToEndpoint("viewLog");
			req.addArgument("log",logId);
			req.addArgument("offset",Integer.toString(offset));
			req.addArgument("count",Integer.toString(records));
			logRecords = aContext.issueRequestForResponse(req);
		}
		
		INKFRequest req=aContext.createRequest("active:xslt");
		req.addArgument("operator", "res:/tools/logging/styleLogView.xsl");
		req.addArgumentByValue("operand",logs);
		if (logRecords!=null) req.addArgumentFromResponse("records",logRecords);
		req.addArgumentByValue("httpParam",param);
		req.setRepresentationClass(aContext.getThisRequest().getRepresentationClass());
		INKFResponseReadOnly resp=aContext.issueRequestForResponse(req);
		INKFResponse respOut=aContext.createResponseFrom(resp);
		respOut.setHeader("WrappedControlPanel", true);
		respOut.setHeader("WrappedControlPanelWide", true);
	}
}
