package tools.scriptplaypen;

import java.util.*;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.urii.ValueSpace;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ISpace;

public class ScriptPlaypen extends StandardAccessorImpl
{
	private static final String PERSISTENCE_BASE="fpds:/tools/scriptplaypen/examples/";
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		IHDSNode params=aContext.source("httpRequest:/params",IHDSNode.class);
		
		String script=(String)params.getFirstValue("script");
		String name=(String)params.getFirstValue("name");
		String type=(String)params.getFirstValue("type");
		String identifier=(String)params.getFirstValue("identifier");
		String spaceString=(String)params.getFirstValue("space");
		if (spaceString!=null && spaceString.length()==0)
		{	spaceString=null;
		}
		
		ISpace space=null;
		if (spaceString!=null)
		{	space=aContext.getKernelContext().getKernel().getSpace(new SimpleIdentifierImpl(spaceString), null, null);
			if (space==null)
			{	throw new NKFException("space not found",spaceString);
			}
		}
		
		Object action=params.getFirstValue("action2");
		if ("execute".equals(action))
		{	doExecute(aContext,script,type,identifier,space);
		}
		else
		{
			HDSBuilder b2 = new HDSBuilder();
			b2.pushNode("data");
			
			String exampleId=(String)params.getFirstValue("example");
			if (exampleId!=null && exampleId.length()>0)
			{	script=aContext.source(exampleId,String.class);
				name=getName(exampleId);
				type=getType(exampleId);
				identifier=exampleId;
				spaceString=null;
			}
			
			String reference=(String)params.getFirstValue("reference");
			if (reference!=null)
			{	//load an external referenced example
				INKFRequest req=aContext.createRequest(reference);
				//IRequestScopeLevel scope=RequestScopeLevelImpl.createOrphanedRootScopeLevel(space, aContext.getKernelContext().getRequestScope());
				//req.setRequestScope(scope);
				req.injectRequestScope(space);
				req.setRepresentationClass(String.class);
				script=(String)aContext.issueRequest(req);
				name=getName(reference);
				type=getType(reference);
				identifier=reference;
			}
			
			if ("save".equals(action))
			{
				identifier=PERSISTENCE_BASE+name+"."+type;
				aContext.sink(identifier, script);
			}
			
			
			b2.addNode("script", script);
			b2.addNode("name", name);
			b2.addNode("type", type);
			b2.addNode("identifier", identifier);
			b2.addNode("space", spaceString);
			
			IHDSNode saves=aContext.source(PERSISTENCE_BASE,IHDSNode.class);
			b2.pushNode("saves");
			for (IHDSNode n : saves.getNodes("/set/identifier"))
			{	String id=n.getValue().toString();
				id=id.replaceAll("%20"," ");
				b2.pushNode("save");
				b2.addNode("identifier",id);
				b2.addNode("type", getType(id));
				b2.addNode("name", getName(id));
				b2.popNode();
			}
			b2.popNode();
			
			
			INKFRequest req=aContext.createRequest("active:xslt");
			req.addArgument("operator", "res:/tools/scriptplaypen/stylePlaypen.xsl");
			req.addArgumentByValue("operand", b2.getRoot());
			INKFResponseReadOnly xsltResponse = aContext.issueRequestForResponse(req);
			
			INKFResponse response=aContext.createResponseFrom(xsltResponse);
			response.setHeader("WrappedControlPanel",true);
		}
	}
	
	private void doExecute(INKFRequestContext aContext, String aScript, String aType, String aIdentifier, ISpace aSpace) throws Exception
	{
		String runtimeIdentifier;
		if (aType.equals("dpml"))
		{	runtimeIdentifier="dpml";
		}
		else if (aType.equals("bsh"))
		{	runtimeIdentifier="beanshell";
		}
		else if (aType.equals("js"))
		{	runtimeIdentifier="javascript";
		}
		else if (aType.equals("py"))
		{	runtimeIdentifier="python";
		}
		else if (aType.equals("gy"))
		{	runtimeIdentifier="groovy";
		}
		else if (aType.equals("rb"))
		{	runtimeIdentifier="ruby";
		}
		else throw new NKFException("Unknown type");
		
		INKFRequest req=aContext.createRequest("active:"+runtimeIdentifier);
		ValueSpace passByValueSpace = new ValueSpace(1);
		String id=aIdentifier;
		if (id.startsWith("fpds:"))
		{	id="res:"+id.substring(5);
		}
		else if (id.length()==0)
		{	id="res:/playpenscript";
		}
		passByValueSpace.put(id, aScript, MetaImpl.EXPIRED);
		req.addArgument("operator",id);
		if (aSpace!=null)
		{	req.injectDurableRequestScope(aSpace);
		}
		req.injectDurableRequestScope(passByValueSpace);
		//req.addArgumentByValue("operator", aScript);
		req.setRepresentationClass(IReadableBinaryStreamRepresentation.class);
		INKFResponseReadOnly resp=aContext.issueRequestForResponse(req);
		aContext.createResponseFrom(resp);
		
	}
	
	private static String getName(String aIdentifier)
	{	int i=aIdentifier.lastIndexOf(".");
		int j=aIdentifier.lastIndexOf("/");
		String name=aIdentifier.substring(j+1, i);
		return name;
	}
	
	private static String getType(String aIdentifier)
	{	int i=aIdentifier.lastIndexOf(".");
		String type=aIdentifier.substring(i+1);
		return type;
	}
}
